from pyvipr.viz import Viz

__all__ = [
    'sp_view',
    'sp_comp_view',
    'sp_comm_view',
    'sp_comm_hierarchy_view',
    'sp_rxns_bidirectional_view',
    'sp_rxns_view',
    'sp_rules_view',
    'sp_rules_fxns_view',
    'sp_rules_mod_view',
    'projected_species_reactions_view',
    'projected_reactions_view',
    'projected_rules_view',
    'projected_species_rules_view',
    'sp_dyn_view',
    'sp_comp_dyn_view',
    'sp_comm_dyn_view',
    'sim_model_dyn_view',
    'sbgn_view'
]


def sp_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interactions between the species in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """

    return Viz(data=model, type_of_viz='sp_view', layout_name=layout_name)


def sp_comp_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interactions between the species in a model.
    The species nodes are grouped by the compartments they belong to.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='sp_comp_view', layout_name=layout_name)


def sp_comm_view(model, layout_name='klay', random_state=None):
    """
    Render a visualization of the interactions between the species in a model.
    The species nodes are grouped by the communities detected by the 
    Louvain algorithm: https://en.wikipedia.org/wiki/Louvain_Modularity.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use
    random_state: int
        Random state seed use by the community detection algorithm

    """
    return Viz(data=model, type_of_viz='sp_comm_view', random_state=random_state, layout_name=layout_name)


def sp_comm_hierarchy_view(model, layout_name='klay', random_state=None):
    """
    Render a visualization of the interactions between the species in a model.
    The species nodes are grouped by the communities detected by the
    Louvain algorithm: https://en.wikipedia.org/wiki/Louvain_Modularity.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use
    random_state: int
        Random state seed use by the community detection algorithm

    """
    return Viz(data=model, type_of_viz='sp_comm_hierarchy_view', random_state=random_state, layout_name=layout_name)


def sp_rxns_bidirectional_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the
    bidirectional reactions.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='sp_rxns_bidirectional_view', layout_name=layout_name)


def sp_rxns_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the
    unidirectional reactions.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='sp_rxns_view', layout_name=layout_name)


def sp_rules_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='sp_rules_view', layout_name=layout_name)


def sp_rules_fxns_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.
    Additionally, rules are grouped by the macros that created them.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='sp_rules_fxns_view', layout_name=layout_name)


def sp_rules_mod_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.
    Additionally, rules are grouped by the modules where they are defined.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='sp_rules_mod_view', layout_name=layout_name)


def projected_species_reactions_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interactions between species in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='projected_species_reactions_view', layout_name=layout_name)


def projected_reactions_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interaction between the reaction in a model

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='projected_reactions_view', layout_name=layout_name)


def projected_rules_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interactions between rules in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='projected_rules_view', layout_name=layout_name)


def projected_species_rules_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return Viz(data=model, type_of_viz='projected_species_rules_view', layout_name=layout_name)


def sbgn_view(model, layout_name='cose-bilkent'):
    return Viz(data=model, type_of_viz='sbgn_view', layout_name=layout_name)


def sp_dyn_view(simulation, sim_idx=0, process='consumption', layout_name='cose-bilkent'):
    """
    Render a dynamic visualization of the simulation

    Parameters
    ----------
    simulation : pysb.SimulationResult
        Simulation result to visualize
    sim_idx : int
        Index of simulation to be visualized
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name : str
        Layout to use

    """
    return Viz(data=simulation, type_of_viz='dynamic_sp_view', layout_name=layout_name,
               process=process, sim_idx=sim_idx)


def sp_comp_dyn_view(simulation, sim_idx=0, process='consumption', layout_name='cose-bilkent'):
    """
    Render a dynamic visualization of the simulation. The species nodes are grouped
    by the compartments they belong to.

    Parameters
    ----------
    simulation: pysb.SimulationResult object
        Simulation result to visualize dynamically
    sim_idx: int
        Index of simulation to be visualized
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name: str
        Layout to use

    """
    return Viz(data=simulation, type_of_viz='dynamic_sp_comp_view', layout_name=layout_name,
               process=process, sim_idx=sim_idx)


def sp_comm_dyn_view(simulation, sim_idx=0, process='consumption', layout_name='klay', random_state=None):
    """
    Render a dynamic visualization of the simulation. The species nodes are grouped
    by the communities detected by the Louvain algorithm

    Parameters
    ----------
    simulation: pysb.SimulationResult object
        Simulation result to visualize dynamically
    sim_idx: int
        Index of simulation to be visualized
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name: str
        Layout to use
    random_state: int
        Random state seed use by the community detection algorithm

    """
    return Viz(data=simulation, type_of_viz='dynamic_sp_comm_view', layout_name=layout_name,
               random_state=random_state, process=process, sim_idx=sim_idx)


def sim_model_dyn_view(model, tspan, param_values=None, type_of_viz='dynamic_view',
                       process='consumption', layout_name='cose-bilkent'):
    """
    Render a dynamic visualization of the model using the tspan and param_values
    passed to the function

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    tspan : vector-like, optional
        Time values over which to simulate. The first and last values define
        the time range.
    param_values : vector-like or dict, optional
        Values to use for every parameter in the model. Ordering is
        determined by the order of model.parameters.
        If passed as a dictionary, keys must be parameter names.
        If not specified, parameter values will be taken directly from
        model.parameters.
    type_of_viz: str
        Type of visualization. It can only be `sp_dyn_view`, `sp_comp_dyn_view`
        or `sp_comm_dyn_view`
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name : str
        Layout name to use

    """
    from pysb.simulator import ScipyOdeSimulator
    from pyvipr.util import dispatch_pysb_files

    model = dispatch_pysb_files(model)
    sim = ScipyOdeSimulator(model, tspan=tspan).run(param_values=param_values)
    return Viz(data=sim, type_of_viz=type_of_viz, process=process, layout_name=layout_name)



import numpy as np
import tellurium
import networkx as nx
from pyvipr.tellurium_viz.static_viz import TelluriumStaticViz
from pyvipr.util_networkx import from_networkx
import pyvipr.util as hf
try:
    import tesbml as libsbml
except ImportError:
    import libsbml


class TelluriumDynamicViz(object):
    """
    class to visualize the dynamics of systems biology models defined in sbml or antimony format

    Parameters
    ----------
    sim_model: tellurium roadrunner
        A roadrunner instance after a simulation
    """
    mach_eps = np.finfo(float).eps

    def __init__(self, sim_model):
        if isinstance(sim_model, tellurium.roadrunner.extended_roadrunner.ExtendedRoadRunner):
            model_sbml = sim_model.getSBML()
            self.doc = libsbml.readSBMLFromString(model_sbml)
            self.model = self.doc.getModel()
        else:
            raise Exception('Model must be a roadrunner instance')

        required_selections = ['time'] + [s.getId() for s in self.model.species] + \
                              [r.getId() for r in self.model.reactions]
        sim_selections = sim_model.selections

        # Check that the simulation selections include all species and reactions rate values
        if not all(x in required_selections for x in sim_selections):
            raise ValueError('To use this visualization you must use '
                             'this simulator selection \n {0}'.format(required_selections))
        self.y = sim_model.getSimulationData()
        self.sp_graph = None
        self.type_viz = 'consumption'

    def dynamic_sp_view(self, type_viz='consumption'):
        """
        Generates a dictionary with the model dynamics data that can be converted in the Cytoscape.js JSON format

        Parameters
        ----------
        type_viz : str
            Type of the dynamic visualization, it can be 'consumption' or 'production'

        Returns
        -------
        dict
            A Dictionary Object with all nodes and edges information that
            can be converted into Cytoscape.js JSON to be visualized
        """
        self.type_viz = type_viz
        self.sp_graph = TelluriumStaticViz(self.model).species_graph()
        self.sp_graph.graph['nsims'] = 1
        self.sp_graph.graph['tspan'] = self.y['time'].tolist()
        self._add_edge_node_dynamics()
        data = from_networkx(self.sp_graph)
        return data

    def matrix_reaction_rates(self):
        rxns_matrix = np.zeros((len(self.model.reactions), len(self.y['time'])))

        # Calculates matrix of bidirectional reaction rates
        for idx, reac in enumerate(self.model.reactions):
            rxns_matrix[idx] = self.y[reac.getId()]
        return rxns_matrix

    def edges_colors_sizes(self):
        """
        This function obtains values for the size and color of the edges in the network.
        The color is a representation of the percentage of flux going through an edge.
        The edge size is a representation of the relative value of the reaction normalized to the maximum value that
        the edge can attain during the whole simulation.

        Returns
        -------
        tuple
            Three dictionaries. The first one contains the information of the edge sizes at all time points.
            The second one contains the information of the edge colors at all time points.
            The third one contains the values of the reaction rates at all time points.
        """
        all_rate_colors = {}
        all_rate_sizes = {}
        all_rate_abs_val = {}

        rxns_matrix = self.matrix_reaction_rates()
        all_products = [[s.getSpecies() for s in rx.getListOfProducts()] for rx in self.model.reactions]
        all_reactants = [[s.getSpecies() for s in rx.getListOfReactants()] for rx in self.model.reactions]

        for sp in self.model.species:
            rxns_idx_reactant = [i for i, rx in enumerate(all_reactants) if sp.getId() in rx]
            rxns_idx_product = [i for i, rx in enumerate(all_products) if sp.getId() in rx]
            repeated_rxn = set(rxns_idx_reactant) & set(rxns_idx_product)
            # Obtain indices of reactions in which sp is only a product
            rxns_idx_product = [c for c in rxns_idx_product if c not in repeated_rxn]

            # Getting arrays to obtain producing and consuming reactions
            rxn_val_pos = np.where(
                np.concatenate((rxns_matrix[rxns_idx_reactant] > 0, rxns_matrix[rxns_idx_product] < 0)),
                rxns_matrix[rxns_idx_reactant + rxns_idx_product], 0)  # Reactants are consumed
            rxn_val_neg = np.where(
                np.concatenate((rxns_matrix[rxns_idx_reactant] < 0, rxns_matrix[rxns_idx_product] > 0)),
                rxns_matrix[rxns_idx_reactant + rxns_idx_product], 0)  # Reactants are produced

            rxn_val_pos = np.abs(rxn_val_pos)
            rxn_val_neg = np.abs(rxn_val_neg)
            # An array with the total of reaction rates at each time point
            rxn_val_pos_total = rxn_val_pos.sum(axis=0)
            rxn_val_neg_total = rxn_val_neg.sum(axis=0)

            # The max and min of the group of reaction rates. It can be used to assign size to edges
            # based in the max and min of the groups of reaction rates across all time points
            # if rxn_val_pos.size != 0:
            #     rxn_val_pos_max = np.amax(rxn_val_pos) + self.mach_eps
            #
            # if rxn_val_neg.size != 0:
            #     rxn_val_neg_max = np.amax(rxn_val_neg) + self.mach_eps

            if self.type_viz == 'consumption':
                sp_rr_dom = rxns_matrix[rxns_idx_reactant]

                for rx_idx, rx in enumerate(rxns_idx_reactant):
                    products = all_products[rx]
                    for p in products:
                        # Normalizing by the total flux in a node
                        # We ignore division by zero and invalid value in less than function warnings
                        # as they are handled later on by converting nan values to 0
                        with np.errstate(divide='ignore', invalid='ignore'):
                            react_rate_color = rxns_matrix[rx] / rxn_val_pos_total
                            rxn_neg_idx = np.where(react_rate_color < 0)
                            react_rate_color[rxn_neg_idx] = (
                                    rxns_matrix[rx][rxn_neg_idx] / rxn_val_neg_total[rxn_neg_idx])
                        np.nan_to_num(react_rate_color, copy=False)
                        rate_colors = hf.f2hex_edges(react_rate_color)

                        rxn_eps = sp_rr_dom[rx_idx] + self.mach_eps  # rxns_matrix[rx] + self.mach_eps
                        react_rate_size = np.zeros(len(rxn_eps))
                        tro_pos_idx = np.where(rxn_eps > 0)[0]
                        tro_neg_idx = np.where(rxn_eps < 0)[0]

                        if tro_neg_idx.size != 0:
                            rxn_neg_max = np.amax(np.abs(rxn_eps[tro_neg_idx]))
                            react_rate_size[tro_neg_idx] = np.array([rxn_neg/rxn_neg_max for
                                                                     rxn_neg in np.abs(rxn_eps[tro_neg_idx])])

                        if tro_pos_idx.size != 0:
                            rxn_pos_max = np.amax(rxn_eps[tro_pos_idx])
                            react_rate_size[tro_pos_idx] = np.array([rxn_pos/rxn_pos_max for
                                                                     rxn_pos in rxn_eps[tro_pos_idx]])

                        rate_sizes = hf.range_normalization(react_rate_size, min_x=0, max_x=1)
                        edges_id = (sp.getId(), p)
                        all_rate_colors[edges_id] = rate_colors
                        all_rate_sizes[edges_id] = rate_sizes.tolist()
                        all_rate_abs_val[edges_id] = rxns_matrix[rx].tolist()

            elif self.type_viz == 'production':
                sp_prr_dom = rxns_matrix[rxns_idx_product]

                for rp_idx, rp in enumerate(rxns_idx_product):
                    reactants = all_reactants[rp]
                    for r in reactants:
                        with np.errstate(divide='ignore', invalid='ignore'):
                            preact_rate_color = rxns_matrix[rp] / rxn_val_neg_total
                            prxns_neg_idx = np.where(preact_rate_color < 0)
                            preact_rate_color[prxns_neg_idx] = (
                                    rxns_matrix[rp][prxns_neg_idx] / rxn_val_pos_total[prxns_neg_idx])
                        np.nan_to_num(preact_rate_color, copy=False)
                        prate_colors = hf.f2hex_edges(preact_rate_color)

                        prxn_eps = sp_prr_dom[rp_idx] + self.mach_eps
                        preact_rate_size = np.zeros(len(prxn_eps))
                        ptro_pos_idx = np.where(prxn_eps > 0)[0]
                        ptro_neg_idx = np.where(prxn_eps < 0)[0]

                        if ptro_neg_idx.size != 0:
                            prxn_neg_max = np.amax(np.abs(prxn_eps[ptro_neg_idx]))
                            preact_rate_size[ptro_neg_idx] = np.array([prxn_neg/prxn_neg_max for
                                                                       prxn_neg in np.abs(prxn_eps[ptro_neg_idx])])

                        if ptro_pos_idx.size != 0:
                            prxn_pos_max = np.amax(prxn_eps[ptro_pos_idx])
                            preact_rate_size[ptro_pos_idx] = np.array([prxn_pos/prxn_pos_max for
                                                                       prxn_pos in prxn_eps[ptro_pos_idx]])

                        prate_sizes = hf.range_normalization(preact_rate_size, min_x=0, max_x=1)
                        edges_id = (r, sp.getId())
                        all_rate_colors[edges_id] = prate_colors
                        all_rate_sizes[edges_id] = prate_sizes.tolist()
                        all_rate_abs_val[edges_id] = rxns_matrix[rp].tolist()
            else:
                raise ValueError('The type of process can only be `consumption` or `production`')
        return all_rate_sizes, all_rate_colors, all_rate_abs_val

    def node_data(self):
        """
        Obtains the species concentration values and the relative concentration
        compared with the maximum concentration across all time points

        Returns
        -------
        tuple
            Two dictionaries. The first one has the species concentration. The
            second one has the relative species concentrations
        """
        node_absolute = {}
        node_relative = {}
        for sp in self.model.species:
            sp_absolute = np.absolute(self.y[sp.getId()])
            sp_relative = (sp_absolute / sp_absolute.max()) * 100
            node_absolute[sp.getId()] = sp_absolute.tolist()
            node_relative[sp.getId()] = sp_relative.tolist()

        # all_nodes_values = pandas.DataFrame(all_rate_colors)
        return node_absolute, node_relative

    def _add_edge_node_dynamics(self):
        """
        Add the edge size and color data as well as node color and values data
        Returns
        -------

        """
        edge_sizes, edge_colors, edge_qtips = self.edges_colors_sizes()
        nx.set_edge_attributes(self.sp_graph, edge_colors, 'edge_color')
        nx.set_edge_attributes(self.sp_graph, edge_sizes, 'edge_size')
        nx.set_edge_attributes(self.sp_graph, edge_qtips, 'edge_qtip')

        node_abs, node_rel = self.node_data()
        nx.set_node_attributes(self.sp_graph, node_abs, 'abs_value')
        nx.set_node_attributes(self.sp_graph, node_rel, 'rel_value')



import networkx as nx
from pyvipr.util_networkx import from_networkx
import pysb
from pysb.bng import generate_equations
from networkx.algorithms import bipartite
import re
import pyvipr.util as hf


class PysbStaticViz(object):
    """
    Class to generate static visualizations of systems biology models

    Parameters
    ----------
    model : pysb.Model
        PySB Model to visualize.
    """

    def __init__(self, model):
        # Need to create a model visualization base and then do independent visualizations: static and dynamic
        self.model = model
        self.graph = None

    @staticmethod
    def merge_nodes(G, nodes, new_node, **attr):
        """
        Merges the selected `nodes` of the graph G into one `new_node`,
        meaning that all the edges that pointed to or from one of these
        `nodes` will point to or from the `new_node`.
        attr_dict and `**attr` are defined as in `G.add_node`.
        """
        G.add_node(new_node, **attr)  # Add the 'merged' node
        newG = G.copy()

        for n1, n2, data in newG.edges(data=True):
            # For all edges related to one of the nodes to merge,
            # make an edge going to or coming from the `new gene`.
            if n1 in nodes:
                G.add_edge(new_node, n2, **data)
            elif n2 in nodes:
                G.add_edge(n1, new_node, **data)

        for n in nodes:  # remove the merged nodes
            G.remove_node(n)

    def sp_view(self):
        """
        Generate a dictionary that contains the species network information

        Parameters
        ----------

        Examples
        --------
        >>> from pysb.examples.earm_1_0 import model
        >>> viz = PysbStaticViz(model)
        >>> data = viz.sp_view()

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, positions) to generate a cytoscapejs network.
        """
        graph = self.species_graph()
        data = from_networkx(graph)
        return data

    def sp_comp_view(self):
        """
        Generate a dictionary that contains the information about the species 
        network. Species are grouped by the compartments they belong to

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes, parent_nodes, edges, positions) to
            generate a cytoscapejs network.
        """
        graph = self.compartments_data_graph()
        data = from_networkx(graph)
        return data

    def compartments_data_graph(self):
        """
        Create a networkx DiGraph. Check for compartments in a model and add 
        the compartments as compound nodes where the species are located

        Returns
        -------
        nx.Digraph
            Graph with model species and compartments
        
        Raises
        ------
        ValueError
            Model has not compartments
        """
        # Check if model has comparments
        if not self.model.compartments:
            raise ValueError('Model has no compartments')
        graph = self.species_graph()
        compartment_nodes = []
        for c in self.model.compartments:
            if c.parent is not None:
                compartment_nodes.append((c.name, dict(parent=c.parent.name)))
            else:
                compartment_nodes.append(c.name)

        graph.add_nodes_from(compartment_nodes, NodeType='compartment')
        # Determining compartment node family tree
        # Determine species compartment
        sp_compartment = {}
        for idx, sp in enumerate(self.model.species):
            monomers = sp.monomer_patterns
            monomers_comp = {m.compartment.name: m.compartment.size for m in monomers}
            smallest_comp = min(monomers_comp, key=monomers_comp.get)
            sp_compartment['s{0}'.format(idx)] = smallest_comp
        nx.set_node_attributes(graph, sp_compartment, 'parent')
        return graph

    def sp_comm_view(self, random_state=None):
        """
        Use the Louvain algorithm https://en.wikipedia.org/wiki/Louvain_Modularity
        for community detection to find groups of nodes that are densely connected.
        It generates the data to create a network with compound nodes that hold the communities.

        Parameters
        ==========
        random_state : int, optional
            Random state seed use by the community detection algorithm, by default None

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        graph = self.communities_data_graph(random_state=random_state)
        data = from_networkx(graph)
        return data

    def sp_comm_hierarchy_view(self, random_state=None):
        """
        Use the Louvain algorithm https://en.wikipedia.org/wiki/Louvain_Modularity
        for community detection to find groups of nodes that are densely connected.
        It generates the data of all the intermediate clusters obtained during the Louvain
        algorithm generate to create a network with compound nodes that hold the communities.

        Parameters
        ==========
        random_state : int, optional
            Random state seed use by the community detection algorithm, by default None

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        graph = self.communities_data_graph(all_levels=True, random_state=random_state)
        data = from_networkx(graph)
        return data

    def communities_data_graph(self, all_levels=False, random_state=None):
        """
        Create a networkx DiGraph. It applies the Louvain algorithm to detect
        communities and add that information to the graph
        
        Parameters
        ----------
        all_levels : bool
            Indicates if the generated graph contains the information about the
            clusters obtained at each iteration of the Louvain algorithm
        random_state : int, optional
            Random state seed use by the community detection algorithm, by default None
        
        Returns
        -------
        nx.DiGraph
            A networkx DiGraph where the nodes have a `parent` property that correspond
            to the community they belong to
        """
        graph = self.species_graph()
        hf.add_communities(graph, all_levels=all_levels, random_state=random_state)

        return graph

    def sp_rxns_bidirectional_view(self):
        """
        Generate a dictionary with the info of a bipartite graph where one set of
        nodes is the model species and the other set is the model bidirectional reactions

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.

        """
        graph = self.sp_rxns_bidirectional_graph()
        data = from_networkx(graph)
        return data

    def sp_rxns_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the unidirectional reactions

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        graph = self.sp_rxns_graph()
        data = from_networkx(graph)
        return data

    def sp_rules_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the model rules
        
        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        rules_graph = self.rules_graph()
        rules_graph = self.graph_merged_pair_edges(rules_graph)
        data = from_networkx(rules_graph)
        return data

    def sp_rules_fxns_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the model rules. Additionally, it adds information of the functions from which
        the rules come from.
        
        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        rules_graph = self.rules_graph()
        rules_graph = self.graph_merged_pair_edges(rules_graph)
        rule_functions = {rule.name: rule._function for rule in self.model.rules}
        unique_functions = set(rule_functions.values())
        nx.set_node_attributes(rules_graph, rule_functions, 'parent')
        rules_graph.add_nodes_from(unique_functions, NodeType='function')
        data = from_networkx(rules_graph)
        return data

    def sp_rules_mod_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the model rules. Additionally, it adds information of the modules from which
        the rules come from.
        
        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        rules_graph = self.rules_graph()
        rules_graph = self.graph_merged_pair_edges(rules_graph)

        # Unique modules used in the model
        unique_modules = [m for m in self.model.modules if not m.startswith('_')]
        # Remove outer model file to not look further modules that are not related to
        # the model
        unique_modules.remove(self.model.name)
        module_parents = {}
        # pysb components have a modules list that starts from the inner frame
        # to the outer frame.
        for module in unique_modules:
            total_parents = []
            for rule in self.model.rules:
                mods = rule._modules
                try:
                    mod_idx = mods.index(module)
                    module_parent = mods[mod_idx + 1]
                except (ValueError, IndexError):
                    continue
                total_parents.append(module_parent)

            total_parents = list(set(total_parents))
            if module in total_parents:
                total_parents.remove(module)

            if len(total_parents) >= 2:
                for idx, p in enumerate(total_parents):
                    module_parents[module + str(idx)] = p
            else:
                module_parents[module] = total_parents[0]

        rules_module = {}
        for rule in self.model.rules:
            m = rule._modules[0]
            if m not in module_parents.keys():
                m_parent = rule._modules[1]
                m_parent_child = list(module_parents.keys())[list(module_parents.values()).index(m_parent)]
                rules_module[rule.name] = m_parent_child
            else:
                rules_module[rule.name] = m

        module_parent_nodes = list(module_parents.keys())
        module_parent_nodes.append(self.model.name)
        rules_graph.add_nodes_from(module_parent_nodes, NodeType='module')
        nx.set_node_attributes(rules_graph, module_parents, 'parent')
        nx.set_node_attributes(rules_graph, rules_module, 'parent')
        data = from_networkx(rules_graph)
        return data

    def sbgn_view(self):
        sbgn_graph = self.sbgn_graph()
        data = from_networkx(sbgn_graph)
        return data

    def projections_view(self, project_to='species_reactions'):
        """
        Project a bipartite graph to the type of node defined in `project to`.
        Generates a dictionary with the info of a graph representing the PySB model.

        Parameters
        ----------
        project_to: str
            Nodes to which the graph is going to be projected. Options are:
            'species_reactions', 'reactions', 'rules', 'species_rules'

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        if project_to == 'species_reactions' or project_to == 'reactions':
            bipartite_graph = self.sp_rxns_bidirectional_graph()
        elif project_to == 'rules' or project_to == 'species_rules':
            bipartite_graph = self.rules_graph()
        else:
            raise ValueError('Projection not valid')

        projected_graph = self.projected_graph(bipartite_graph, project_to)
        data = from_networkx(projected_graph)
        return data

    def projected_species_reactions_view(self):
        return self.projections_view('species_reactions')

    def projected_reactions_view(self):
        return self.projections_view('reactions')

    def projected_rules_view(self):
        return self.projections_view('rules')

    def projected_species_rules_view(self):
        return self.projections_view('species_rules')

    def _sp_initial(self, sp):
        """
        Get initial condition of a species
        Parameters
        ----------
        sp: pysb.ComplexPattern, pysb species

        Returns
        -------

        """
        sp_0 = 0
        for spInitial in self.model.initials:
            if spInitial.pattern.is_equivalent_to(sp):
                sp_0 = spInitial.value.value
                break
        return sp_0

    def species_graph(self):
        """
        Creates a nx.DiGraph graph of the model species interactions

        Returns
        -------
        nx.Digraph 
            Graph that has the information for the visualization of the model
        """
        generate_equations(self.model)
        sp_graph = nx.DiGraph(name=self.model.name)

        # TODO: there are reactions that generate parallel edges that are not taken into account because netowrkx
        # digraph only allows one edge between two nodes
        for idx, sp in enumerate(self.model.species):
            species_node = 's%d' % idx
            color = "#2b913a"
            # color species with an initial condition differently
            if len([s.pattern for s in self.model.initials if s.pattern.is_equivalent_to(sp)]):
                color = "#aaffff"
            # Setting the information about the node
            node_data = dict(label=parse_name(sp),
                             background_color=color,
                             shape='ellipse',
                             NodeType='species',
                             spInitial=self._sp_initial(sp))
            sp_graph.add_node(species_node, **node_data)

        for reaction in self.model.reactions_bidirectional:
            reactants = set(reaction['reactants'])
            products = set(reaction['products'])
            if reaction['reversible']:
                attr_reversible = {'source_arrow_shape': 'diamond', 'target_arrow_shape': 'triangle',
                                   'source_arrow_fill': 'hollow'}
            else:
                attr_reversible = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle',
                                   'source_arrow_fill': 'filled'}

            for s in reactants:
                for p in products:
                    self._r_link_species(sp_graph, s, p, **attr_reversible)
        return sp_graph

    @staticmethod
    def _r_link_species(graph, s, r, **attrs):
        """
        Links two nodes in a species graph
        Parameters
        ----------
        s : int
            Source node
        r: int
            Target node
        attrs: dict
            Other attributes for edges

        Returns
        -------

        """

        nodes = ('s{0}'.format(s), 's{0}'.format(r))
        attrs.setdefault('arrowhead', 'normal')
        graph.add_edge(*nodes, **attrs)

    def sp_rxns_bidirectional_graph(self):
        """
        Creates a bipartite nx.DiGraph graph where one set of nodes is the model species
        and the other set is the model bidirectional reactions.

        Returns
        -------
        nx.Digraph 
            Graph that has the information for the visualization of the model
        """
        generate_equations(self.model)
        graph = nx.DiGraph(name=self.model.name, graph={'rankdir': 'LR'})
        for i, cp in enumerate(self.model.species):
            species_node = 's%d' % i
            slabel = parse_name(cp)
            color = "#2b913a"
            # color species with an initial condition differently
            if len([s.pattern for s in self.model.initials if s.pattern.is_equivalent_to(cp)]):
                color = "#aaffff"
            graph.add_node(species_node,
                           label=slabel,
                           shape="ellipse",
                           background_color=color,
                           NodeType='species',
                           spInitial=self._sp_initial(cp),
                           bipartite=0)
        for j, reaction in enumerate(self.model.reactions_bidirectional):
            reaction_node = 'r%d' % j
            rule = self.model.rules.get(reaction['rule'][0])
            graph.add_node(reaction_node,
                           label=reaction_node,
                           shape="roundrectangle",
                           background_color="#d3d3d3",
                           NodeType='reaction',
                           kf=rule.rate_forward.value,
                           kr=rule.rate_reverse.value if rule.rate_reverse else 'None',
                           bipartite=1)
            reactants = set(reaction['reactants'])
            products = set(reaction['products'])
            modifiers = reactants & products
            reactants = reactants - modifiers
            products = products - modifiers
            attr_reversible = {'source_arrow_shape': 'triangle', 'target_arrow_shape': 'triangle',
                               'source_arrow_fill': 'hollow'} if reaction['reversible'] \
                else {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle', 'source_arrow_fill': 'filled'}
            for s in reactants:
                self._r_link_bipartite(graph, s, j, **attr_reversible)
            for s in products:
                self._r_link_bipartite(graph, s, j, _flip=True, **attr_reversible)
            for s in modifiers:
                attr_modifiers = {'target_arrow_shape': 'diamond'}
                self._r_link_bipartite(graph, s, j, **attr_modifiers)
        return graph

    def sp_rxns_graph(self):
        """
        Creates a bipartite nx.DiGraph graph where one set of nodes is the model species
        and the other set is the model unidirectional reactions.

        Returns
        -------
        nx.Digraph 
            Graph that has the information for the visualization of the model
        """
        generate_equations(self.model)
        graph = nx.DiGraph(name=self.model.name, graph={'rankdir': 'LR'})
        for i, cp in enumerate(self.model.species):
            species_node = 's%d' % i
            slabel = parse_name(self.model.species[i])
            color = "#2b913a"
            # color species with an initial condition differently
            if len([s.pattern for s in self.model.initials if s.pattern.is_equivalent_to(cp)]):
                color = "#aaffff"
            graph.add_node(species_node,
                           label=slabel,
                           shape="ellipse",
                           background_color=color,
                           NodeType='species',
                           spInitial=self._sp_initial(cp),
                           bipartite=0)
        for i, reaction in enumerate(self.model.reactions):
            reaction_node = 'r%d' % i
            rule = self.model.rules.get(reaction['rule'][0])
            graph.add_node(reaction_node,
                           label=reaction_node,
                           shape="roundrectangle",
                           background_color="#d3d3d3",
                           NodeType='reaction',
                           kf=rule.rate_forward.value if not reaction['reverse'][0] else 'None',
                           kr=rule.rate_reverse.value if reaction['reverse'][0] else 'None',
                           bipartite=1)
            reactants = set(reaction['reactants'])
            products = set(reaction['products'])
            modifiers = reactants & products
            reactants = reactants - modifiers
            products = products - modifiers
            attr_edges = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle', 'source_arrow_fill': 'filled'}
            for s in reactants:
                self._r_link_bipartite(graph, s, i, **attr_edges)
            for s in products:
                self._r_link_bipartite(graph, s, i, _flip=True, **attr_edges)
            for s in modifiers:
                self._r_link_bipartite(graph, s, i, **attr_edges)
        return graph

    def sbgn_graph(self):
        import re
        from pysb.pattern import Pattern, Name
        from pysb import Rule, Parameter
        graph = nx.DiGraph(name=self.model.name, style='sbgn')
        all_cp = []
        cp_to_delete = []
        rules_for_nodes = []
        # Obtain complex patterns from rules, save the complexes formed during catalysis to remove them
        # as they don't appear in the SBGN standard, and save the rules that are not related to catalysis
        for r in self.model.rules:
            if re.match('catalyze_.+_to_.+_.+', r.name):
                cp_to_delete.append(r.reactant_pattern.complex_patterns[0])
                all_cp.append([r.product_pattern.complex_patterns[1]])
            elif re.match('bind_.+_.+_to_.+', r.name): # and r._function == 'catalyze':
                all_cp.append(r.reactant_pattern.complex_patterns)
                all_cp.append(r.product_pattern.complex_patterns)
            else:
                all_cp.append(r.reactant_pattern.complex_patterns)
                all_cp.append(r.product_pattern.complex_patterns)
                rules_for_nodes.append(r)
        # For each complex that is formed during the catalysis, we obtain the bind and catalyze rules to generate
        # the new rules that would give us the SBGN standard visualization
        for cpd in cp_to_delete:
            bind_cat = self.model.rules.filter(Pattern(cpd))
            bind = bind_cat.filter(Name('^bind'))
            cat = bind_cat.filter(Name('^catalyze'))
            par_aux = Parameter('bla', 1, _export=False)
            rule_aux1 = Rule('production_cat_1', bind[0].reactant_pattern.complex_patterns[1] >>
                             cat[0].product_pattern.complex_patterns[1], par_aux, _export=False)
            rule_aux2 = Rule('catalysis_cat_2', bind[0].reactant_pattern.complex_patterns[0] >>
                             None, par_aux, _export=False)
            rule_aux = {'production_cat_1': rule_aux1, 'catalysis_cat_2': rule_aux2}
            rules_for_nodes.append(rule_aux)
        # obtain unique complex patterns to add to the network
        all_cp = [item for sublist in all_cp for item in sublist]
        unique_cp = []
        for i in all_cp:
            if not in_cp_list(i, unique_cp) and not in_cp_list(i, cp_to_delete):
                unique_cp.append(i)
        cp_encode = [(unique_cp[idx], idx) for idx in range(len(unique_cp))]

        for cp, cp_idx in cp_encode:
            cp_node = 's{0}'.format(cp_idx)
            slabel = parse_name(cp)
            cp_length = len(cp.monomer_patterns)
            if cp_length == 1:
                cp_class = 'macromolecule'
                mp_node = 'mp1_{0}'.format(cp_node)
                mp = cp.monomer_patterns[0]
                state_variables = []
                for s, v in mp.site_conditions.items():
                    if isinstance(v, str):
                        state = {'variable': s, 'value': v}
                        state_variables.append({'id': mp_node, 'class': 'state variable', 'state': state})
                node_data = {'label': mp.monomer.name, 'class': cp_class,
                             'NodeType': 'macromolecule', 'stateVariables': state_variables, "unitsOfInformation": []}
                graph.add_node(cp_node, **node_data)

            elif cp_length == 2:
                cp_class = 'complex'
                node_data = {'label': slabel, 'class': cp_class, 'stateVariables': [],
                             'NodeType': 'complex', "unitsOfInformation": []}
                graph.add_node(cp_node, **node_data)
                for idx, mp in enumerate(cp.monomer_patterns):
                    mp_node = 'mp2_{0}_{1}'.format(cp_node, idx)
                    state_variables = []
                    for s, v in mp.site_conditions.items():
                        if isinstance(v, str):
                            state = {'variable': s, 'value': v}
                            state_variables.append({'id': mp_node, 'class': 'state variable', 'state': state})
                    node_data_mp = {'label': mp.monomer.name, 'stateVariables': state_variables,
                                    'class': 'macromolecule', 'unitsOfInformation': [], 'parent': cp_node}
                    graph.add_node(mp_node, **node_data_mp)

            elif cp_length >= 3:
                cp_class = 'macromolecule multimer'

        for r_idx, rule in enumerate(rules_for_nodes):
            rule_node = 'r%d' % r_idx
            node_data = {'label': rule_node, 'class': 'process', 'NodeType': 'process'}
            graph.add_node(rule_node, **node_data)
            if isinstance(rule, dict):
                for rname, rcat in rule.items():
                    if rname.startswith('catalysis_cat'):
                        sbgn_class = 'catalysis'
                    elif rcat.is_reversible:
                        sbgn_class = 'production'
                    else:
                        sbgn_class = 'consumption'
                    reactants = rcat.reactant_pattern.complex_patterns
                    products = rcat.product_pattern.complex_patterns

                    for s in reactants:
                        reactant_node = get_cp_idx(s, cp_encode)
                        attr_edges = {'class': sbgn_class, "cardinality": 0,
                                      "portSource": "s{0}".format(reactant_node), "portTarget": rule_node}
                        self._r_link_bipartite(graph, reactant_node, r_idx, **attr_edges)
                    for s in products:
                        p_node = get_cp_idx(s, cp_encode)
                        attr_edges = {'class': 'production', "cardinality": 0,
                                      "portSource": rule_node, "portTarget": "s{0}".format(p_node)}
                        self._r_link_bipartite(graph, p_node, r_idx, _flip=True, **attr_edges)

            else:
                reactants = rule.reactant_pattern.complex_patterns
                products = rule.product_pattern.complex_patterns
                for s in reactants:
                    reactant_node = get_cp_idx(s, cp_encode)
                    attr_edges = {'class': 'consumption', "cardinality": 0,
                                  "portSource": "s{0}".format(reactant_node), "portTarget": rule_node}
                    self._r_link_bipartite(graph, reactant_node, r_idx, **attr_edges)
                for s in products:
                    p_node = get_cp_idx(s, cp_encode)
                    attr_edges = {'class': 'production', "cardinality": 0,
                                  "portSource": rule_node, "portTarget": "s{0}".format(p_node)}
                    self._r_link_bipartite(graph, p_node, r_idx, _flip=True, **attr_edges)
        return graph

    def rules_graph(self):
        """
        Creates a bipartite nx.DiGraph graph where one set of nodes is the model species
        and the other set is the model rules.

        Returns
        -------
        nx.Digraph
            Graph that has the information for the visualization of the model
        """
        graph = self.sp_rxns_graph()
        # Merge reactions into the rules that they come from
        nodes2merge = self.merge_reactions2rules()
        node_attrs = {'shape': 'roundrectangle', 'background_color': '#ff4c4c',
                      'NodeType': 'rule', 'bipartite': 1}
        for rule_info, rxns in nodes2merge.items():
            rule = self.model.rules.get(rule_info[0])
            node_attrs['kf'] = rule.rate_forward.value
            node_attrs['kr'] = rule.rate_reverse.value if rule.rate_reverse else 'None'
            node_attrs['label'] = rule_info[0]
            node_attrs['index'] = 'rule' + str(rule_info[1])
            self.merge_nodes(graph, rxns, rule_info[0], **node_attrs)
        return graph

    def projected_graph(self, graph, project_to='species'):
        """
        Project a bipartite graph into one of the sets of nodes

        Parameters
        ----------
        graph: nx.DiGraph
            a networkx bipartite graph
        project_to: str
            One of the following options `species_reactions`, `species_rules`,
            `reactions`, `rules`

        Returns
        -------
        nx.DiGraph
            Projected graph
        """
        if project_to == 'species_reactions' or project_to == 'species_rules':
            nodes = {n for n, d in graph.nodes(data=True) if d['bipartite'] == 0}
        elif project_to == 'reactions' or project_to == 'rules':
            nodes = {n for n, d in graph.nodes(data=True) if d['bipartite'] == 1}
        else:
            raise ValueError('Projection not valid')
        projected_graph = bipartite.projected_graph(graph, nodes)
        projected_graph = self.graph_merged_pair_edges(projected_graph)

        return projected_graph

    @staticmethod
    def graph_merged_pair_edges(graph):
        """
        Merges pair of edges that are reversed
        
        Parameters
        ----------
        graph: nx.DiGraph
            The networkx directed graph whose pairs of edges ((u, v), (v, u)) are going to be merged
        
        Returns
        -------
        nx.Digraph 
            Graph that has the information for the visualization of the model
        """
        edges_to_delete = []
        edges_attributes = {}
        for edge in graph.edges():
            if edge in edges_to_delete:
                continue
            if graph.has_edge(*edge[::-1]):
                attr_reversible = {'source_arrow_shape': 'triangle', 'target_arrow_shape': 'triangle',
                                   'source_arrow_fill': 'hollow'}
                edges_attributes[edge] = attr_reversible
                edges_to_delete.append(edge[::-1])
            else:
                attr_irreversible = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle',
                                     'source_arrow_fill': 'filled'}
                edges_attributes[edge] = attr_irreversible
        graph.remove_edges_from(edges_to_delete)
        nx.set_edge_attributes(graph, edges_attributes)
        return graph

    def merge_reactions2rules(self):
        """
        Merges the model reactions into each of the rules from which the reactions come form.
        
        Returns
        -------
        dict
            Dictionary whose keys are tuples of rule name and rule index and the values
            are the reactions that are generated by each rule
        """
        rxn_per_rule = {}
        for r_idx, rule in enumerate(self.model.rules):
            rxns = []
            for i, rxn in enumerate(self.model.reactions):
                if rxn['rule'][0] == rule.name:
                    rxns.append('r{0}'.format(i))
            rxn_per_rule[(rule.name, r_idx)] = rxns
        return rxn_per_rule

    @staticmethod
    def _r_link_bipartite(graph, s, r, **attrs):
        """
        Links two nodes in a species graph
        Parameters
        ----------
        s : int
            Source node
        r: int
            Target node
        attrs: dict
            Other attributes for edges

        Returns
        -------

        """

        nodes = ('s{0}'.format(s), 'r{0}'.format(r))
        if attrs.get('_flip'):
            del attrs['_flip']
            nodes = nodes[::-1]
        # attrs.setdefault('arrowhead', 'normal')
        graph.add_edge(*nodes, **attrs)


def in_cp_list(cp, cp_list):
    if [s for s in cp_list if s.is_equivalent_to(cp)]:
        return True
    else:
        return False


def get_cp_idx(cp, cp_list):
    idx = None
    for c, c_idx in cp_list:
        if c.is_equivalent_to(cp):
            idx = c_idx
            break
    return idx


def parse_name(spec):
    """
    Function that writes short names of the species to name the nodes.
    It counts how many times a monomer_pattern is present in the complex pattern an its states
    then it takes only the monomer name and its state to write a shorter name to name the nodes.

    Parameters
    ----------
    spec : pysb.ComplexPattern
        Name of species to parse

    Returns
    -------
    Parsed name of species
    """
    m = spec.monomer_patterns
    lis_m = []
    name_counts = {}
    parsed_name = ''
    for i in range(len(m)):
        sp_name = str(m[i]).partition('(')[0]
        sp_comp = str(m[i]).partition('** ')[-2:]
        sp_comp = ''.join(sp_comp)
        sp_states = re.findall(r"['\"](.*?)['\"]", str(m[i])) # Matches strings between quotes
        sp_states = [s.lower() for s in sp_states]
        # tmp_2 = re.findall(r"(?<=\').+(?=\')", str(m[i]))
        if not sp_states and not sp_comp:
            lis_m.append(sp_name)
        else:
            sp_states.insert(0, sp_name)
            sp_states.insert(0, sp_comp)
            sp_states.reverse()
            lis_m.append(''.join(sp_states))
    for name in lis_m:
        name_counts[name] = lis_m.count(name)

    for sp, counts in name_counts.items():
        if counts == 1:
            parsed_name += sp + '-'
        else:
            parsed_name += str(counts) + sp + '-'
    return parsed_name[:len(parsed_name) - 1]


from pyvipr.util_networkx import from_networkx


class NetworkViz(object):
    def __init__(self, network):
        self.network = network

    def network_static_view(self):
        self.network.graph['name'] = ''
        json = from_networkx(self.network)
        return json

    def dynamic_network_view(self):
        # This is necessary because the javascript part chooses to render
        # a static or dynamic visualization depending on the name
        # of the function.
        data = from_networkx(self.network)
        return data



from pyvipr.viz import Viz

__all__ = [
    'nx_graph_view',
    'nx_graph_dyn_view',
]


def nx_graph_view(graph, layout_name='cose'):
    """
    Render a networkx Graph or DiGraph
    Parameters
    ----------
    graph: nx.Graph or nx.DiGraph
        Graph to render
    layout_name: str
        Layout to use
    Returns
    -------
    """
    return Viz(data=graph, type_of_viz='network_static_view', layout_name=layout_name)


def nx_graph_dyn_view(graph, tspan, node_rel=None, node_tip=None, edge_colors=None, edge_sizes=None,
                      edge_tips=None, layout_name='cose'):
    """
    Render a dynamic visualization of a networkx graph
    Parameters
    ----------
    graph: nx.DiGraph or nx.Graph
    tspan: vector-like, optional
        Time values over which to simulate. The first and last values define
    node_rel: dict
        A dictionary where the keys are the node ids and the values are
        lists that contain (0-100) values that are represented in a
        pie chart within the node
    node_tip: dict
        A dictionary where the keys are the node ids and the values are
        lists that contain any value that can be accessed
        as a tooltip in the rendered network
    edge_colors: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any hexadecimal color value that are
        represented in the edge colors
    edge_sizes: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any numerical value that are
        represented in the edge size
    edge_tips: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any value that can be accessed
        as a tooltip in the rendered network
    layout_name: str
        Layout to use
    """
    from pyvipr.util_networkx import network_dynamic_data

    network_dynamic_data(graph, tspan, node_rel, node_tip, edge_colors, edge_sizes,
                         edge_tips)

    return Viz(data=graph, type_of_viz='dynamic_network_view', layout_name=layout_name)


from pyvipr.viz import Viz

__all__ = [
    'sp_view',
    'sp_comm_view',
    'sp_dyn_view'
]


def sp_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interactions between the species in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """

    return Viz(data=model, type_of_viz='sp_view', layout_name=layout_name)


def sp_dyn_view(simulation, process='consumption', layout_name='cose-bilkent'):
    """
    Render a dynamic visualization of the simulation

    Parameters
    ----------
    simulation : pysb.SimulationResult
        Simulation result to visualize
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name : str
        Layout to use

    """
    return Viz(data=simulation, type_of_viz='dynamic_sp_view', layout_name=layout_name,
               process=process)


def sp_comm_view(model, layout_name='klay', random_state=None):
    """
    Render a visualization of the interactions between the species in a model.
    The species nodes are grouped by the communities detected by the
    Louvain algorithm: https://en.wikipedia.org/wiki/Louvain_Modularity.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use
    random_state: int
        Random state seed use by the community detection algorithm

    """
    return Viz(data=model, type_of_viz='sp_comm_view', random_state=random_state, layout_name=layout_name)



import networkx as nx
from six import string_types
import tellurium
from pyvipr.util_networkx import from_networkx
import pyvipr.util as hf
import math
try:
    import tesbml as libsbml
except ImportError:
    import libsbml


class TelluriumStaticViz(object):
    """
    Class to generate static visualization of sbml models from tellurium
    """

    def __init__(self, model):
        if isinstance(model, tellurium.roadrunner.extended_roadrunner.ExtendedRoadRunner):
            model_sbml = model.getSBML()
            self.doc = libsbml.readSBMLFromString(model_sbml)
            self.model = self.doc.getModel()
        elif isinstance(model, string_types):
            self.doc = libsbml.readSBMLFromString(model)
            self.model = self.doc.getModel()
        elif isinstance(model, libsbml.SBMLDocument):
            self.doc = model
            self.model = self.doc.getModel()
        elif isinstance(model, libsbml.Model):
            self.model = model
        else:
            raise Exception('SBML Input is not valid')
        self.graph = None

    def species_graph(self):
        """
        Creates a graph of the model species interactions
        Returns
        -------
        nx.Digraph
            Graph that has the information for the visualization of the model
        """
        sp_graph = nx.DiGraph(name='Tellurium model')
        for s in self.model.species:
            if s.getName():
                label = s.getName()
            else:
                label = s.getId()
            if s.getInitialConcentration() == 0 or math.isnan(s.getInitialConcentration()):
                color = "#2b913a"
            else:
                color = "#aaffff"

            node_data = dict(label=label,
                             background_color=color,
                             shape='ellipse',
                             NodeType='species',
                             spInitial=s.getInitialConcentration())
            sp_graph.add_node(s.getId(), **node_data)

        for r in self.model.reactions:

            # edges
            reactants = r.getListOfReactants()
            products = r.getListOfProducts()
            if r.getReversible():
                attr_edge = {'source_arrow_shape': 'diamond', 'target_arrow_shape': 'triangle',
                             'source_arrow_fill': 'hollow', 'arrowhead': 'normal'}
            else:
                attr_edge = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle',
                             'source_arrow_fill': 'filled', 'arrowhead': 'normal'}
            for s in reactants:
                for p in products:
                    sp_graph.add_edge(s.getSpecies(), p.getSpecies(), **attr_edge)

        return sp_graph

    def sp_view(self):
        """
        Generate a dictionary that contains the species network information


        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, positions) to generate a cytoscapejs network.
        """
        graph = self.species_graph()
        data = from_networkx(graph)
        return data

    def communities_data_graph(self, all_levels=False, random_state=None):
        """
        Create a networkx DiGraph. It applies the Louvain algorithm to detect
        communities and add that information to the graph

        Parameters
        ----------
        all_levels : bool
            Indicates if the generated graph contains the information about the
            clusters obtained at each iteration of the Louvain algorithm
        random_state : int, optional
            Random state seed use by the community detection algorithm, by default None

        Returns
        -------
        nx.DiGraph
            A networkx DiGraph where the nodes have a `parent` property that correspond
            to the community they belong to
        """
        graph = self.species_graph()
        hf.add_communities(graph, all_levels=all_levels, random_state=random_state)
        return graph

    def sp_comm_view(self, random_state=None):
        """
        Use the Louvain algorithm https://en.wikipedia.org/wiki/Louvain_Modularity
        for community detection to find groups of nodes that are densely connected.
        It generates the data to create a network with compound nodes that hold the communities.

        Parameters
        ==========
        random_state : int, optional
            Random state seed use by the community detection algorithm, by default None

        Returns
        -------
        dict
            A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
            contains all the information (nodes,edges, parent nodes, positions) to generate
            a cytoscapejs network.
        """
        graph = self.communities_data_graph(random_state=random_state)
        data = from_networkx(graph)
        return data



from networkx import Graph, DiGraph
import os
import sys


def data_to_json(value, widget):
    """
    Generate a json file from the data passed to the widget
    Parameters
    ----------
    value: pysb.Model, pysb.SimulationResult, str
        Value passed to the widget that is going to be visualized
    widget: Widget
        Widget instance

    Returns
    -------

    """

    if is_pysb_model(value):
        from pyvipr.pysb_viz.static_viz import PysbStaticViz

        viz = PysbStaticViz(value)
        jsondata = static_data(viz, widget)
        return jsondata

    elif is_pysb_sim(value):
        from pyvipr.pysb_viz.dynamic_viz import PysbDynamicViz

        viz = PysbDynamicViz(value, widget.sim_idx)
        jsondata = dynamic_data(viz, widget)
        return jsondata

    elif isinstance(value, str):
        try:
            from pysb.importers.sbml import model_from_sbml, model_from_biomodels
            from pysb.importers.bngl import model_from_bngl
        except ImportError:
            raise Exception('PySB must be installed to visualize models from files')
        from pyvipr.pysb_viz.static_viz import PysbStaticViz
        file_extension = os.path.splitext(value)[1]
        if file_extension == '.bngl':
            model = model_from_bngl(value)
        elif file_extension in ['.sbml', '.xml']:
            model = model_from_sbml(value)
        elif value.startswith('BIOMD'):
            model = model_from_biomodels(value)
        # elif file_extension == '.ka':
        #     subprocess.run(['truml', '-k', value])
        #     bngl_model_path = re.sub('ka', 'bngl', value)
        #     model = model_from_bngl(bngl_model_path)
        #     os.remove(bngl_model_path)

        else:
            raise ValueError('Format not supported')
        viz = PysbStaticViz(model)
        jsondata = static_data(viz, widget)
        return jsondata

    elif is_tellurium_model(value):
        if widget.type_of_viz.startswith('dynamic'):
            from pyvipr.tellurium_viz.dynamic_viz import TelluriumDynamicViz
            viz = TelluriumDynamicViz(value)
            jsondata = dynamic_data(viz, widget)
        else:
            from pyvipr.tellurium_viz.static_viz import TelluriumStaticViz
            viz = TelluriumStaticViz(value)
            jsondata = static_data(viz, widget)
        return jsondata

    elif isinstance(value, (DiGraph, Graph)):
        from pyvipr.networkx_viz.network_viz import NetworkViz
        viz = NetworkViz(value)
        jsondata = getattr(viz, widget.type_of_viz)()
        return jsondata

    elif is_ecell_model(value):
        try:
            from pysb.importers.sbml import model_from_sbml
        except ImportError:
            raise Exception('PySB must be installed to visualize models from files')
        from pyvipr.pysb_viz.static_viz import PysbStaticViz
        import tempfile
        ecell4 = sys.modules['ecell4']
        f_sbml = tempfile.NamedTemporaryFile(suffix='.sbml')
        # In ecell4 species don't have initial conditions as attributes. Hence, the
        # initial conditions are passed as a dictionary to the save_sbml function.
        # If no initial conditions are passed ecell4 sets the initial condition of the
        # species to 0, and PySB throws an error when the initial condition of all the species
        # are zero. For visualization purposes we then set the initial conditions to 1.
        y0 = {sp.serial(): 1 for sp in value.list_species()}
        ecell4.util.ports.save_sbml(f_sbml.name, value, y0=y0)
        model = model_from_sbml(f_sbml.name)
        viz = PysbStaticViz(model)
        jsondata = static_data(viz, widget)
        return jsondata

    elif is_pysces_model(value):
        try:
            from pysb.importers.sbml import model_from_sbml
        except ImportError:
            raise Exception('PySB must be installed to visualize models from files')
        from pyvipr.pysb_viz.static_viz import PysbStaticViz
        import tempfile
        # Note: Importing a pysces model to sbml doesn't work in python 3.7
        pysces = sys.modules['pysces']
        f_sbml = tempfile.NamedTemporaryFile(suffix='.xml')
        pysces.interface.writeMod2SBML(value, f_sbml.name)
        model = model_from_sbml(f_sbml.name)
        viz = PysbStaticViz(model)
        jsondata = static_data(viz, widget)
        return jsondata

    else:
        raise TypeError('Only pysb Model, pysb SimulationResult, tellurium Model, '
                        'PySCeS Model, and networkx graphs are supported')


def static_data(viz_obj, w):
    try:
        if w.type_of_viz in ['sp_comm_view', 'sp_comm_hierarchy_view']:
            rs = w.random_state
            jsondata = getattr(viz_obj, w.type_of_viz)(random_state=rs)
        else:
            jsondata = getattr(viz_obj, w.type_of_viz)()
    except AttributeError:
        raise AttributeError('Type of static visualization not defined')
    return jsondata


def dynamic_data(viz_object, w):
    process = w.process
    try:
        if w.type_of_viz == 'dynamic_sp_comm_view':
            rs = w.random_state
            jsondata = getattr(viz_object, w.type_of_viz)(random_state=rs, type_viz=process)
        else:
            jsondata = getattr(viz_object, w.type_of_viz)(type_viz=process)
    except AttributeError:
        raise AttributeError('Type of visualization not defined')
    return jsondata


def is_pysb_model(obj):
    if 'pysb.core' in sys.modules:
        return isinstance(obj, sys.modules['pysb.core'].Model)
    else:
        return False


def is_pysb_sim(obj):
    if 'pysb.simulator' in sys.modules:
        return isinstance(obj, sys.modules['pysb.simulator'].SimulationResult)
    else:
        return False


def is_tellurium_model(obj):
    if 'tellurium' in sys.modules:
        return isinstance(obj, sys.modules['tellurium'].roadrunner.extended_roadrunner.ExtendedRoadRunner)
    else:
        return False


def is_pysces_model(obj):
    if 'pysces' in sys.modules:
        return isinstance(obj, sys.modules['pysces'].PyscesModel.PysMod)
    else:
        return False


def is_ecell_model(obj):
    if 'ecell4_base' in sys.modules:
        return isinstance(obj, sys.modules['ecell4_base'].core.NetworkModel)
    else:
        return False
